from .functional import SublevelsetPersistence
import torch.nn as nn

class SublevelsetDiagram(nn.Module):
    def __init__(self, X, maxdim=0, reduction_flags=()):
        super(SublevelsetDiagram, self).__init__()
        self.X = X
        self.maxdim = maxdim
        self.reduction_flags = reduction_flags
        self.PD = SublevelsetPersistence()

    def forward(self, f):
        return self.PD.apply(f, self.X, self.maxdim, *self.reduction_flags)



import torch
import numpy as np

def dgms_tensor_list(ReducedCC, maxHomdim):
    """
    Persistent Diagrams to tensor list
    Return a list of tensors, where each tensor is a Persistent Diagrams at a dimension
    Note:
    1. We also return zero length Bar for gradient computation consideration
    2. The everlasting persistence bars will have death index as 
        64-bit unsigned maximum integer, and so we set it to be -1 here. 
    3. The everlasting persistence bars will have death values as float('inf'), 
        so you can use `p.death() == float('inf')` to find the it.
        
    -Inputs:
        ReducedCC - Reduced Chain Complex in bats
        maxHomdim - maximum homology dimension

    -Outputs:
        dgms: a list of PD tensors
        bdinds: birth-death indices (used to find gradient later)
    """
    dgms = []
    bdinds = []
    for i in range(maxHomdim + 1):
        bd_pair, bd_inds = ReducedCC.persistence_pairs_vec(i)
        # bd_pair: a list of birth-death pairs with 
        # length 2 * number of bd pairs. 
        # also note that it includes zeros length ones
        bd_inds = np.array(bd_inds)
        bd_pair = np.array(bd_pair)
        bd_inds[bd_inds == 0xFFFFFFFFFFFFFFFF] = -1 # take care of bats.NO_IND

        # convert to tensor
        bd_pair = torch.tensor(bd_pair.reshape(-1,2), requires_grad = True)
        bd_inds = torch.tensor(bd_inds.reshape(-1,2), requires_grad = False, dtype=torch.long)

        # add it
        dgms.append(bd_pair)
        bdinds.append(bd_inds)

    return dgms, bdinds



import bats
import torch
from torch.autograd import Function
import scipy.spatial.distance as distance
from .util import dgms_tensor_list

def compute_y_gradient(X, F, R, imap, grad_dgms):
    '''
    -Input: 
        X: original data points
        F: Filtration built on X
        R: Reduced Chain Complex computed on F
        imap: inverse map that maps the birth or death index of a persistence pair 
            to the critical edge that creates or destroys it.
        grad_dgms: gradient with respect to PDs that is accumulated from the last layer.

    -Output: 
        y_grad: gradient values that is to be updated to X.
    
    '''
    y_grad = torch.zeros_like(X)
    scplex = F.complex()

    for i, grad_dgm in enumerate(grad_dgms):
        ps = R.persistence_pairs(i) # persistence pair at dim i
        
        for ind, gd in enumerate(grad_dgm):
            # non-zero gradient
            if not torch.equal(gd, torch.tensor([0.,0.], dtype=torch.double)):
                # find correponding critical simplex index in filtration
                p = ps[ind]
                d = p.dim() # homology dimension

                # Birth
                # get indices of two vertices of the (birth) edge
                bi = p.birth_ind() # index of birth edge

                [birth_vertex1, birth_vertex2] = scplex.get_simplex(1, bi)

                # compute gradient on y now for birth index
                dx = X[birth_vertex1] - X[birth_vertex2]
                if torch.norm(dx) == 0:
                    dx = 0
                else:
                    dx /= torch.norm(dx)
                y_grad[birth_vertex1] = gd[0] * dx
                y_grad[birth_vertex2] = - gd[0] * dx

                # Death (avoid infinite death)
                if p.death() != float('inf') and p.death_ind() <= len(imap[d+1])-1:
                    # p.death_ind() now is the index of an triangle that destroys H1
                    # we need to map it to the critical edge that creates it
                    di = imap[d+1][p.death_ind()]

                    # get index of two vertices of the (death) edge
                    [death_vertex1, death_vertex2] = scplex.get_simplex(1, di)

                    # compute gradient on y now for death index
                    dx = X[death_vertex1] - X[death_vertex2]
                    dx /= torch.norm(dx)
                    y_grad[death_vertex1] = gd[1] * dx
                    y_grad[death_vertex2] = - gd[1] * dx


    return y_grad


class RipsDiagram(Function):
    """
    Compute Rips complex persistence using point coordinates
    forward inputs:
        y - N x D torch.float tensor of coordinates (original data points)
        maxdim - maximum homology dimension
        reduction_flags - optional reduction flags for BATS.
    """
    @staticmethod
    def forward(ctx, y, maxdim, *reduction_flags):
        # number of arguments should match the return of backward function

        ynp = y.detach().numpy()
        DX = distance.squareform(distance.pdist(ynp))
        rX = bats.enclosing_radius(bats.Matrix(DX))
        # maixmum complex dimension = maximum homology dimension + 1
        F, imap = bats.LightRipsFiltration_extension(bats.Matrix(DX), rX , maxdim+1)
        R = bats.reduce(F, bats.F2(), *reduction_flags)

        # store device
        device = y.device
        ctx.device = device
        ycpu = y.cpu()

        # save data coordinates for backward
        ctx.save_for_backward(ycpu)
        # return persistent diagrams with death and birth values
        dgms, bdinds = dgms_tensor_list(R, maxdim)

        # use context variable `ctx` to store information that will be used for backward
        ctx.R = R
        ctx.filtration = F
        ctx.imap = imap
        ctx.reduction_flags = reduction_flags
        return tuple(dgms)

    @staticmethod
    def backward(ctx, *grad_dgms):
        """
        Input: 
            grad_dgms: a Tensor containing the gradient of the loss
            with respect to the output of this layer (i.e., gradient w.r.t. PDs) 
            and now we need to compute the gradient of the loss 
            with respect to the first argument in input (original data points).
        """

        # find returned gradient, which is the same size as dgms in forward function
        grad_dgm = [gd.cpu() for gd in grad_dgms]

        # find the gradient on y with the same shape as y
        device = ctx.device
        ycpu, = ctx.saved_tensors
        R = ctx.R
        F = ctx.filtration
        imap = ctx.imap

        grad_y = compute_y_gradient(ycpu, F, R, imap, grad_dgms)

        # backward only to the first argument in inputs: y 
        ret = [grad_y.to(device), None] 
        # 'None' here means that we need to match the forward function arguments
        # and `maxdim` and `*reduction_flags` cannot do gradient-descent
        ret.extend([None for f in ctx.reduction_flags])
        return tuple(ret)



# Distance based on 2 Pesistence Diagrams:
# 1. BottleneckDistance
# 2. WassersteinDistance
import torch.nn as nn
from .functional import BottleneckDistance, WassersteinDistance
from .poly_feat import remove_zero_bars

# `from . import` is called Intra-package References 
# see in https://docs.python.org/3/tutorial/modules.html#intra-package-references

class BottleneckLayer(nn.Module):
    def __init__(self):
        super(BottleneckLayer, self).__init__()
        self.D = BottleneckDistance()

    def forward(self, dgm0, dgm1):
        dgm0 = remove_zero_bars(dgm0)
        dgm1 = remove_zero_bars(dgm1)

        return self.D.apply(dgm0, dgm1)

class WassersteinLayer(nn.Module):
    def __init__(self):
        super(WassersteinLayer, self).__init__()
        self.D = WassersteinDistance()

    def forward(self, dgm0, dgm1):
        dgm0 = remove_zero_bars(dgm0)
        dgm1 = remove_zero_bars(dgm1)

        return self.D.apply(dgm0, dgm1)



import torch
from torch.autograd import Function
from .util import dgms_tensor_list
import bats

class SublevelsetPersistence(Function):
    """
    Compute persistent homology of a

    forward inputs:
        X - a bats SimplicialComplex or LightSimplicialComplex
        f - funcion on the vertex set
    """
    @staticmethod
    def forward(ctx, f, X, maxdim, *reduction_flags):
        ctx.dtype = f.dtype
        f = f.detach().numpy()
        ctx.shape = f.shape

        vals, imap = bats.lower_star_filtration(X, f)
        ctx.imap = imap
        F = bats.Filtration(X, vals)
        R = bats.reduce(F, bats.F2(), *reduction_flags)
        dgms, ctx.binds = dgms_tensor_list(R, maxdim)
        ctx.reduction_flags = reduction_flags

        return tuple(dgms)


    @staticmethod
    def backward(ctx, *grad_dgm):
        device = grad_dgm[0].device

        grad_dgm = [gd.detach().flatten().tolist() for gd in grad_dgm]
        bdinds = [bd.detach().flatten().tolist() for bd in ctx.binds]
        grad_f = bats.lower_star_backwards(grad_dgm, bdinds, ctx.imap)
        grad_f = torch.tensor(grad_f, dtype=ctx.dtype)

        ret = [grad_f.to(device), None, None]
        ret.extend([None for f in ctx.reduction_flags])

        return tuple(ret)



from .functional import RipsDiagram

import torch.nn as nn

class RipsLayer(nn.Module):
    """
    Define a Rips persistence layer that will use the Rips Diagram function
    Inpute:
        maxdim : maximum homology dimension (default=0)
        reduction_flags : PH computation options from bats
            see details in: 
            https://bats-tda.readthedocs.io/en/latest/tutorials/Rips.html#Algorithm-optimization  
    """
    def __init__(self, maxdim = 0, reduction_flags=()):
        super(RipsLayer, self).__init__()
        self.maxdim = maxdim
        self.PD = RipsDiagram()
        self.reduction_flags = reduction_flags

    def forward(self, x):
        xnp = x.cpu().detach().numpy() # convert to numpy array
        dgms = self.PD.apply(x, self.maxdim, *self.reduction_flags)
        return dgms



#!/usr/bin/env python
# Notes helpful for write setup.py
# How to write a setup.py file
# https://godatadriven.com/blog/a-practical-guide-to-using-setup-py/

# Use `pip install -e .` to intall in editable mode
# So you would use this when trying to install a package locally, 
# most often in the case when you are developing it on your system. 
# It will just link the package to the original location, 
# basically meaning any changes to the original package 
# would reflect directly in your environment.

from setuptools import setup
import setuptools

with open('README.md') as f:
    long_description = f.read()

setup(name='torch-tda',
      version='0.0.1',
      description='Automatic differentiation for topological data analysis',
      long_description=long_description,
      long_description_content_type="text/markdown",
      author='Brad Nelson, Yuan Luo',
      author_email='bradnelson@uchicago.edu, yuanluo@uchicago.edu',
      url='https://github.com/CompTop/torch-tda',
      project_urls={
        "Documentation": "https://torch-tda.readthedocs.io/en/latest/",
      },
      license='MIT',
      # package_dir={"": "torch_tda"},
      # packages=setuptools.find_packages(where="torch_tda"),
      packages=['torch_tda', 'torch_tda.nn', 'torch_tda.nn.functional'],
      include_package_data=True,
      install_requires=[
        'numpy',
        'bats-tda',
        'persim>=0.3.1',
        'scipy'
      ],
      python_requires='>=3.7',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'Intended Audience :: Education',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Scientific/Engineering :: Mathematics',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
      ],
      keywords='topological data analysis, automatic differentiation, persistent homology'
     )



# Bottleneck distance

import bats
import torch
from torch.autograd import Function
from persim import bottleneck
import numpy as np

class BottleneckDistance(Function):
    """
    Compute bottleneck distance between two persistence diagrams

    forward inputs:
        dgm0 - N x 2 torch.float tensor of birth-death pairs
        dgm1 - M x 2 torch.float tensor of birth-death pairs
    """
    @staticmethod
    def forward(ctx, dgm0, dgm1):
        ctx.dtype = dgm0.dtype
        d0 = dgm0.detach().numpy()
        d1 = dgm1.detach().numpy()
        n0 = len(dgm0)
        ctx.n0 = n0
        n1 = len(dgm1)
        ctx.n1 = n1

        dist, match = bottleneck(d0, d1, matching=True)

        imax = np.argmax(match[:,2])
        i0, i1, dm = match[imax]
        i0, i1 = int(i0), int(i1)
        # TODO check for -1 as index

        ctx.i0 = i0
        ctx.i1 = i1

        d01 = torch.tensor(d0[i0] - d1[i1], dtype=ctx.dtype)
        ctx.d01 = d01
        dist01 = np.linalg.norm(d0[i0] - d1[i1], np.inf)
        ctx.indmax = np.argmax(np.abs(d0[i0] - d1[i1]))

        return torch.tensor(dist01, dtype=ctx.dtype)

    @staticmethod
    def backward(ctx, grad_dist):
        n0 = ctx.n0
        n1 = ctx.n1
        i0 = ctx.i0
        i1 = ctx.i1
        d01 = ctx.d01

        gd0 = torch.zeros(n0, 2, dtype=ctx.dtype)
        gd1 = torch.zeros(n1, 2, dtype=ctx.dtype)


        gd0[i0, ctx.indmax] = np.sign(d01[ctx.indmax]) * grad_dist
        gd1[i1, ctx.indmax] = -np.sign(d01[ctx.indmax]) * grad_dist

        return gd0, gd1



# Wasserstein distance

import bats
import torch
from torch.autograd import Function
from persim import wasserstein
import numpy as np

class WassersteinDistance(Function):
    """
    Compute bottleneck distance between two persistence diagrams

    forward inputs:
        dgm0 - N x 2 torch.float tensor of birth-death pairs
        dgm1 - M x 2 torch.float tensor of birth-death pairs
    """
    @staticmethod
    def forward(ctx, dgm0, dgm1):
        d0 = dgm0.detach().numpy()
        d1 = dgm1.detach().numpy()
        n0 = len(dgm0)
        ctx.n0 = n0
        n1 = len(dgm1)
        ctx.n1 = n1

        dist, match = wasserstein(d0, d1, matching=True)

        # matching indices
        ctx.match0 = np.array(match[:,0], dtype=int)
        ctx.match1 = np.array(match[:,1], dtype=int)
        ctx.offsets = np.empty((len(match), 2), dtype=float)

        # construct unit offset vector
        inds01 = ctx.match0 > -1
        ctx.offsets[inds01] = dgm0[ctx.match0[inds01]] - dgm1[ctx.match1[inds01]]

        # handle -1 in match0
        # note these are matched with dgm1
        ctx.offsets[ctx.match0 == -1] = [1, -1] # drive apart birth, death

        # handle -1 in match1
        # note these are matched with dgm0
        ctx.offsets[ctx.match1 == 1] =[-1, 1] # negative of above, drive apart birth, death

        # normalize
        ctx.offsets = ctx.offsets / np.linalg.norm(ctx.offsets, axis=1).reshape(-1,1)
        ctx.offsets = torch.tensor(ctx.offsets, dtype=torch.double)
        ctx.match0 = torch.tensor(ctx.match0, dtype=torch.long)
        ctx.match1 = torch.tensor(ctx.match1, dtype=torch.long)

        return torch.tensor(dist)

    @staticmethod
    def backward(ctx, grad_dist):

        gd0 = torch.zeros(ctx.n0, 2, dtype=torch.double)
        gd1 = torch.zeros(ctx.n1, 2, dtype=torch.double)

        gd0[ctx.match0] = ctx.offsets[ctx.match0] * grad_dist
        gd1[ctx.match1] = -ctx.offsets[ctx.match1] * grad_dist

        return gd0, gd1



# The following code is based on code originally obtained from the topology layer
# pacakge (https://github.com/bruel-gabrielsson/TopologyLayer) which is issued
# under the following license
#
#
# MIT License
#
# Copyright (c) 2019 Rickard Brüel-Gabrielsson, Bradley J. Nelson, Anjan Dwaraknath, Primoz Skraba
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import torch
import torch.nn as nn
import numpy as np

def remove_zero_bars(dgm):
    """
    remove zero bars from diagram
    """
    inds = dgm[:,0] != dgm[:,1]
    return dgm[inds,:]


def remove_infinite_bars(dgm, issub):
    """
    remove infinite bars from diagram
    """
    if issub:
        inds = dgm[:, 1] != np.inf
        return dgm[inds,:]
    else:
        inds = dgm[:, 1] != -np.inf
        return dgm[inds,:]

def get_start_end(dgm, issublevel):
    """
    get start and endpoints of barcode pairs
    input:
        dgminfo - Tuple consisting of diagram tensor and bool
            bool = true if diagram is of sub-level set type
            bool = false if diagram is of super-level set type
    output - start, end tensors of diagram
    """
    if issublevel:
        # sub-level set filtration e.g. Rips
        start, end = dgm[:,0], dgm[:,1]
    else:
        # super-level set filtration
        end, start = dgm[:,0], dgm[:,1]
    return start, end


def get_raw_barcode_lengths(dgm, issublevel):
    """
    get barcode lengths from barcode pairs
    no filtering
    """
    start, end = get_start_end(dgm, issublevel)
    lengths = end - start
    return lengths


def get_barcode_lengths(dgm, issublevel):
    """
    get barcode lengths from barcode pairs
    filter out infinite bars
    """
    lengths = get_raw_barcode_lengths(dgm, issublevel)
    # remove infinite and irrelevant bars
    lengths[lengths == np.inf] = 0
    lengths[lengths != lengths] = 0
    return lengths


class SumBarcodeLengths(nn.Module):
    """
    Layer that sums up lengths of barcode in persistence diagram
    ignores infinite bars, and padding
    Options:
        dim - bardocde dimension to sum over (defualt 0)
    forward input:
        (dgms, issub) tuple, passed from diagram layer
    """
    def __init__(self, dim=0):
        super(SumBarcodeLengths, self).__init__()
        self.dim=dim

    def forward(self, dgminfo):
        dgms, issublevel = dgminfo
        lengths = get_barcode_lengths(dgms[self.dim], issublevel)

        # return the sum of the barcode lengths
        return torch.sum(lengths, dim=0)


def get_barcode_lengths_means(dgm, issublevel):
    """
    return lengths and means of barcode
    set irrelevant or infinite to zero
    """
    start, end = get_start_end(dgm, issublevel)
    lengths = end - start
    means = (end + start)/2
    # remove infinite and irrelvant bars
    means[lengths == np.inf] = 0 # note this depends on lengths
    means[lengths != lengths] = 0
    lengths[lengths == np.inf] = 0
    lengths[lengths != lengths] = 0
    return lengths, means


class BarcodePolyFeature(nn.Module):
    """
    applies function
    sum length^p * mean^q
    over lengths and means of barcode
    parameters:
        dim - homology dimension to work over
        p - exponent for lengths
        q - exponent for means
        remove_zero = Flag to remove zero-length bars (default=True)
    """
    def __init__(self, dim, p, q, remove_zero=True):
        super(BarcodePolyFeature, self).__init__()
        self.dim = dim
        self.p = p
        self.q = q
        self.remove_zero = remove_zero

    def forward(self, dgms):
        issublevel = True
        dgm = dgms[self.dim]
        if self.remove_zero:
            dgm = remove_zero_bars(dgm)
        lengths, means = get_barcode_lengths_means(dgm, issublevel)

        return torch.sum(torch.mul(torch.pow(lengths, self.p), torch.pow(means, self.q)))


def pad_k(t, k, pad=0.0):
    """
    zero pad tensor t until dimension along axis is k
    if t has dimension greater than k, truncate
    """
    lt = len(t)
    if lt > k:
        return t[:k]
    if lt < k:
        fillt = torch.tensor(pad * np.ones(k - lt), dtype=t.dtype)
        return torch.cat((t, fillt))
    return t


class TopKBarcodeLengths(nn.Module):
    """
    Layer that returns top k lengths of persistence diagram in dimension
    inputs:
        dim - homology dimension
        k - number of lengths
    ignores infinite bars and padding
    """
    def __init__(self, dim, k):
        super(TopKBarcodeLengths, self).__init__()
        self.k = k
        self.dim = dim

    def forward(self, dgminfo):
        dgms, issublevel = dgminfo
        lengths = get_barcode_lengths(dgms[self.dim], issublevel)

        # sort lengths
        sortl, indl = torch.sort(lengths, descending=True)

        return pad_k(sortl, self.k, 0.0)


class PartialSumBarcodeLengths(nn.Module):
    """
    Layer that computes a partial sum of barckode lengths
    inputs:
        dim - homology dimension
        skip - skip this number of the longest bars
    ignores infinite bars and padding
    """
    def __init__(self, dim, skip):
        super(PartialSumBarcodeLengths, self).__init__()
        self.skip = skip
        self.dim = dim

    def forward(self, dgminfo):
        dgms, issublevel = dgminfo
        lengths = get_barcode_lengths(dgms[self.dim], issublevel)

        # sort lengths
        sortl, indl = torch.sort(lengths, descending=True)

        return torch.sum(sortl[self.skip:])
